#================================#
# MulPro                         #
#================================# 
#
# Author: Max Hantke
# Email: maxhantke@gmail.com

import numpy, time, sys, inspect

import logging
logger = logging.getLogger(__name__)

log_and_raise_error = lambda logger, message: log(logger, message, lvl="ERROR", exception=RuntimeError, rollback=2)
log_warning = lambda logger, message: log(logger, message, lvl="WARNING", exception=None, rollback=2)
log_info = lambda logger, message: log(logger, message, lvl="INFO", exception=None, rollback=2)
log_debug = lambda logger, message: log(logger, message, lvl="DEBUG", exception=None, rollback=2)

def log(logger, message, lvl, exception=None, rollback=1):
    logcalls = {"ERROR": logger.error,
                "WARNING": logger.warning,
                "DEBUG": logger.debug,
                "INFO": logger.info}
    if lvl not in logcalls:
        print "%s is an invalid logger level." % lvl
        sys.exit(1)
    logcall = logcalls[lvl]
    # This should maybe go into a handler
    if (logger.getEffectiveLevel() >= logging.INFO) or rollback is None:
        # Short output
        msg = "%s" % message
    else:
        # Detailed output only in debug mode
        func = inspect.currentframe()
        for r in range(rollback):
            # Rolling back in the stack, otherwise it would be this function
            func = func.f_back
        code = func.f_code
        msg = "%s\n\t=> in \'%s\' function \'%s\' [%s:%i]" % (message,
                                                              func.f_globals["__name__"],
                                                              code.co_name, 
                                                              code.co_filename, 
                                                              code.co_firstlineno)
    logcall("%s:\t%s" % (lvl,msg))
    if exception is not None:
        raise exception(message)
        
def log_execution_time(logger):
    def st_time(func):
        def st_func(*args, **keyArgs):
            t1 = time.time()
            r = func(*args, **keyArgs)
            t2 = time.time()
            try:
                filename = inspect.getsourcefile(func)
                module = inspect.getmodule(func)
                line = inspect.getsourcelines(func)[1]
                loc = "\'%s\' [%s:%i]" % (func.__name__,
                                         filename,
                                         line)
            except TypeError:
                loc = "\'%s\'" % func.__name__
            msg = "Execution time = %.4f sec\n\t=> in function %s" % (t2 - t1,loc)
            log(logger, msg, "DEBUG", exception=None, rollback=None)
            return r        
        return st_func
    return st_time




#================================#
# MulPro                         #
#================================# 
#
# Author: Max Hantke
# Email: maxhantke@gmail.com

import numpy as np
import multiprocessing
import time

import logging
logger = logging.getLogger(__name__)

import log
from log import log_debug, log_info, log_warning, log_and_raise_error

WARNING_AFTER_JOB_DURATION_SEC = 30

import traceback, sys

def _worker_call(worker,pipe):
    try:
        work_package = pipe.recv()
        while work_package is not None:
            t = time.time()
            result = worker(work_package)
            result["_time"] = time.time() - t
            pipe.send(result)
            work_package = pipe.recv()
    except:
        # Put all exception text into an exception and raise that
        raise Exception("".join(traceback.format_exception(*sys.exc_info())))
    
def mulpro(Nprocesses, worker, getwork, logres=None):
    multiprocessing.log_to_stderr(logger.getEffectiveLevel())
    pipes_end_host = list(np.zeros(Nprocesses))
    pipes_end_worker = list(np.zeros(Nprocesses))
    processes = list(np.zeros(Nprocesses))
    processes_job_sent = np.zeros(Nprocesses, dtype=np.bool)
    processes_job_time = np.zeros(Nprocesses, dtype=np.float)

    for i in range(Nprocesses):
        pipes_end_host[i], pipes_end_worker[i] =  multiprocessing.Pipe()
        processes[i] = multiprocessing.Process(target=_worker_call,
                                               args=(worker,pipes_end_worker[i],) )
        processes[i].start()
        log_debug(logger, "Process %i started" % i)

    end = False

    N_done = 0

    t_start = time.time()
    
    # Send out initial jobs
    for i in range(Nprocesses):
        work_package = getwork()
        if work_package is not None:
            pipes_end_host[i].send(work_package)
            processes_job_sent[i] = True
            processes_job_time[i] = time.time()
            log_debug(logger, "Initial job for process %i sent" % i)
        else:
            end = True
            break

    err = False
    
    while not (end and not processes_job_sent.any()) and not err:
        if end:
            log_debug(logger, "Waiting for processes %s to finish" % str(np.arange(Nprocesses)[processes_job_sent]))

        i_processes = range(Nprocesses)
        for i in i_processes:
            waiting_counter = 0
            while not processes[i].is_alive():
                time.sleep(5)
                log_debug(logger, "Process %i: Cannot reach process" % (i))
                waiting_counter += 5
                if waiting_counter > 60:
                    log_and_raise_error(logger, "Process %i died!" % i)
            job_sent = processes_job_sent[i]
            job_duration = time.time()-processes_job_time[i]
            log_debug(logger, "Process %i: job sent = %i (%f sec ago)" % (i,job_sent,job_duration))
            if job_sent:
                log_debug(logger, "Polling process %i" % i)
                if pipes_end_host[i].poll():
                    result = pipes_end_host[i].recv()
                    processes_job_sent[i] = False
                    N_done += 1
                    t_log = time.time()
                    if logres != None:
                        logres(result)
                    t_log = time.time() - t_log
                    log_info(logger, "Datarate %.1f Hz; job %i; process %i/%i; work %.2f sec / logging %.2f sec" % (N_done/(time.time()-t_start),N_done,i,Nprocesses,result["_time"],t_log))
                    if not end:
                        work_package = getwork()
                        if work_package is None:
                            end = True
                        else:
                            pipes_end_host[i].send(work_package)
                            processes_job_sent[i] = True
                            processes_job_time[i] = time.time()
                    else:
                        continue
                else:
                    if job_duration > WARNING_AFTER_JOB_DURATION_SEC:
                        log_warning(logger, "Process %i has not returned from work call for %s sec" % (i,job_duration))
                        #if job_duration > 100:
                        #    log_warning(logger, "Process %i has not returned from work call for more than 1000 sec. Abort." % (i))
                        #    err = True
                        #    break


        time.sleep(0.01)

    for i in range(Nprocesses):
        pipes_end_host[i].send(None)
        processes[i].join(10)
        pipes_end_host[i].close()




#!/usr/bin/env python

from warnings import warn

try:
    import multiprocessing
    multiprocessing_av = True
except ImportError:
    multiprocessing_av = False
        
if multiprocessing_av:
        
    from setuptools import setup

    setup(name='h5writer',
          version='0.1.0',
          description='Running any worker function in parallel and producing output in a separate process',
          author='Max F. Hantke',
          author_email='maxhantke@gmail.com',
          url='https://github.com/mhantke/mulpro',
          #install_requires=['numpy', 'h5py', 'mpi4py>=2.0.0'],
          packages = ['mulpro'],
          #package_dir={'h5writer':'src'},
    )   
    
else:
    
    print 100*"*"

    print "\tThe multiprocessing package cannot be imported! Please install multiprocessing."

    print 100*"*"

